

import pandas as pd
from pdcopyist.src.core import Proxy
from typing import Callable, Iterable
from pdcopyist.src.cusFuns.core.UIModl import Content


class AbcArgs(object):
    """
    docstring
    """
    pass

    def __init__(self, title, default=None, required=True, to_content_overwrite: Callable = None) -> None:
        self.title = title
        self.default = default
        self.required = required
        self.var_name: str = None
        self.to_content_overwrite = to_content_overwrite

    def set_var_name(self, var_name) -> 'AbcArgs':
        self.var_name = var_name
        return self

    def to_content(self) -> Content:
        raise NotImplementedError

    def to_rule(self):
        raise NotImplementedError


class ColumnSelect(AbcArgs):

    def __init__(self, title, default=None, required=True) -> None:
        super().__init__(title, default=default, required=required)

    def to_content(self) -> Content:
        ct = Content(self.title, 'select', self.var_name,
                     source='data_columns', defaultValue=self.default)
        return ct

    def to_rule(self):
        return [{'required': self.required, 'message': '必需选择一列', 'trigger': 'change'}]


class Input(AbcArgs):

    def __init__(self, title, placeholder=None, default=None, required=True) -> None:
        super().__init__(title, default=default, required=required)
        self.placeholder = placeholder

    def to_content(self) -> Content:
        if self.to_content_overwrite:
            return self.to_content_overwrite()

        ct = Content(self.title, 'input', self.var_name,
                     defaultValue=self.default)
        return ct

    def to_rule(self):
        return [{'required': self.required, 'message': '请输入内容', 'trigger': 'blur'}]


class Select(AbcArgs):

    def __init__(self, title, source, default=None, required=True) -> None:
        super().__init__(title, default=default, required=required)

        if isinstance(source, Iterable):
            source = [{'value': i, 'text': str(v)}
                      for i, v in enumerate(source)]
        self.source = source

    def to_content(self) -> Content:
        if self.to_content_overwrite:
            return self.to_content_overwrite()

        ct = Content(self.title, 'select', self.var_name,
                     source=self.source, defaultValue=self.default)
        return ct

    def to_rule(self):
        return [{'required': self.required, 'message': '必需选择一项', 'trigger': 'change'}]


class ColumnWithTypeSelect(AbcArgs):

    def __init__(self, title, default=None, required=True) -> None:
        super().__init__(title, default=default, required=required)

    def get_value(self, index: int):
        df: pd.DataFrame = Proxy.ProxyManager.get().get_last_df_cache()
        col = str(df.columns[index])
        return col

    def to_content(self) -> Content:
        df: pd.DataFrame = Proxy.ProxyManager.get().get_last_df_cache()
        cols = [f'{name}({t})' for name, t in zip(df.columns, df.dtypes)]

        source = [{'value': i, 'text': str(v)}
                  for i, v in enumerate(cols)]

        ct = Content(self.title, 'select', self.var_name,
                     source=source, defaultValue=self.default)
        return ct

    def to_rule(self):
        return [{'required': self.required, 'message': '请输入内容', 'trigger': 'blur'}]


class Switch(AbcArgs):

    def __init__(self, title, default=None, required=True) -> None:
        super().__init__(title, default=default, required=required)

    def to_content(self) -> Content:
        if self.to_content_overwrite:
            return self.to_content_overwrite()

        ct = Content(self.title, 'switch', self.var_name,
                     defaultValue=self.default)
        return ct

    def to_rule(self):
        return [{'required': self.required, 'message': '请选中一项', 'trigger': 'change'}]




from flask import Blueprint
from pdcopyist.src.core.Proxy import ProxyManager
from pdcopyist.src.controllers import common

app = Blueprint('query', __name__)


@app.route('/', methods=['post'])
def query():
    param = common.get_data_from_post()
    px = ProxyManager.get()
    px.query(param['query'])
    return common.df2json(px.get_df_data())



from pdcopyist.src.cusFuns.core.DecoratorFuns import dt_handle_func, dt_args, dt_source_code
from pdcopyist.src.cusFuns.core import Args as ty
import pandas as pd


m_types_select = ['str', 'int']
m_col_with_type_select = ty.ColumnWithTypeSelect('列：')


def generate_code(*args, **kwargs):
    col = f"'{m_col_with_type_select.get_value(kwargs['col'])}'"
    type_value = f"'{m_types_select[kwargs['type']]}'"
    return f'''df[{col}] = df[{col}].astype({type_value})'''


@dt_source_code(generate_code)
@dt_handle_func('列类型转换')
@dt_args(col=m_col_with_type_select, type=ty.Select('类型：', source=m_types_select))
def change_column_type(df: pd.DataFrame, col: str, type: str):
    df[col] = df[col].astype(type)
    return df




from pdcopyist.src.helper.utils import json_converter
import flask
from flask import request, Blueprint
import pandas as pd
from pdcopyist.src.core.Proxy import ProxyManager
from pdcopyist.src.controllers import common

app = Blueprint('command', __name__)


@app.route('/remove', methods=['post'])
def remove_cmd():
    param = common.get_data_from_post()
    id = param['id']

    px = ProxyManager.get()
    px.remove_cmd(id)
    return common.df2json(px.get_df_data())


@app.route('/')
def get_cmds():
    cmds = ProxyManager.get().get_cmds()
    res = flask.json.dumps(cmds, default=json_converter)
    return res


@app.route('/code')
def get_py_code():
    ret = ProxyManager.get().to_code()
    return ret





from pdcopyist.src.controllers import (
    cusFuns_controller, query_controller, table_controller,
    command_controller, groupby_controller, loaddata_controller)
from flask import Flask
import flask
from flask import request
import traceback
import sys


from pdcopyist.src.cusFuns.core.FunsPool import auto_register


import webbrowser
import os

import pathlib
base_path = pathlib.Path(__file__).parent
auto_register(base_path, 'src/cusFuns')


app = Flask(__name__,
            template_folder='./web/templates',
            static_folder='./web/static')


app.register_blueprint(cusFuns_controller.app, url_prefix='/api/cus_fun')
app.register_blueprint(query_controller.app, url_prefix='/api/query')
app.register_blueprint(table_controller.app, url_prefix='/api/table')

app.register_blueprint(command_controller.app, url_prefix='/api/cmd')
app.register_blueprint(groupby_controller.app, url_prefix='/api/groupby')
app.register_blueprint(loaddata_controller.app, url_prefix='/api/data_source')


@app.errorhandler(Exception)
def error_handler(ex):
    exc_type, exc_value, exc_traceback_obj = sys.exc_info()
    return {
        'type':
        'error',
        'code':
        traceback.format_exception(exc_type,
                                   exc_value,
                                   exc_traceback_obj,
                                   limit=2),
        'message':
        repr(ex)
    }


@app.route('/')
def index():
    return flask.render_template('index.html')


def run():

    port = 5551

    # The reloader has not yet run - open the browser
    if not os.environ.get("WERKZEUG_RUN_MAIN"):
        webbrowser.open_new(f'http://localhost:{port}/')

    # Otherwise, continue as normal
    app.run(host="localhost", port=port)


def find_arg(args, flag, return_value=True):
    try:
        idx = args.index(flag)

        if return_value:
            return True, args[idx+1]
        return idx >= 0, None
    except ValueError:
        pass

    return False, None


def copy_bat():
    import pathlib
    import shutil
    src_path = pathlib.Path(__file__).parent / 'startup_win.bat'
    shutil.copyfile(src_path, 'startup_win.bat')


def main():
    args = sys.argv

    # has, dir = find_arg(args, '-dir')
    # if has:
    #     os.chdir(dir)

    has, _ = find_arg(args, '-c', return_value=False)
    if has:
        copy_bat()
        return

    run()


if __name__ == '__main__':
    main()





from typing import Dict, List

from pdcopyist.src.helper.utils import CanJson


class Content(CanJson):
    def __init__(self, title: str, type: str, var_name: str, source=None, defaultValue=None) -> None:
        self.title = title
        self.type = type
        self.var_name = var_name
        self.source = source
        self.defaultValue = defaultValue


class FunModel(CanJson):

    def __init__(self, fun_name: str) -> None:
        self.uid: str = None
        self.function_name = fun_name
        self.contents: List[Content] = []
        self.rules = {}

    def add_content(self, content: Content):
        self.contents.append(content)


class InputRet(object):

    def __init__(self, var_name: str, input: object) -> None:
        self.var_name = var_name
        self.input = input


class CusFunInputRet(object):

    @staticmethod
    def from_dict(data: Dict) -> 'CusFunInputRet':
        pass
        ret = CusFunInputRet()
        ret.uid = data['uid']

        cts = [InputRet(c['content']['var_name'], c['input'])
               for c in data['contents']]

        ret.contents = cts
        return ret

    def __init__(self) -> None:
        self.contents: List[InputRet] = []
        self.uid: str = ''




import flask
from flask import request, Blueprint
from pdcopyist.src.core.Proxy import ProxyManager
from pdcopyist.src.helper.utils import json_converter
from pdcopyist.src.controllers import common

app = Blueprint('cus_fun', __name__)


@app.route('/', methods=['post'])
def cus_fun():
    param = common.get_data_from_post()
    data = param['data']

    px = ProxyManager.get()

    px.run_cus_fun(data)

    return common.df2json(px.get_df_data())


@app.route('/desc')
def get_cus_funcs_desc():
    # auto_register(base_path , 'src/cusFuns')
    ret = ProxyManager.get().get_cus_funcs_desc()
    ret = flask.json.dumps(ret, default=json_converter)
    return ret


@app.route('/model', methods=['get'])
def get_cus_funcs_ui_model():

    uid = str(request.args.get('uid'))

    ret = ProxyManager.get().get_ui_model(uid)
    ret = flask.json.dumps(ret, default=json_converter)
    return ret


@app.route('/input', methods=['post'])
def get_cus_funcs_input():
    param = common.get_data_from_post()
    data = param['data']

    px = ProxyManager.get()

    px.input_cus_funcs(data)
    return common.df2json(px.get_df_data())



from pdcopyist.src.cusFuns.core.DecoratorFuns import dt_handle_func, dt_args
from pdcopyist.src.cusFuns.core import Args as ty
import pandas as pd


@dt_handle_func('删除列')
@dt_args(org_name=ty.ColumnSelect('待删除的列：'))
def remove_cols(df: pd.DataFrame, org_name: str):
    return df.drop(columns=org_name)



import json
from pdcopyist.src.helper.utils import json_converter
from pdcopyist.src.dataModel.pandasCmd import PandasCmd
import flask
import pandas as pd


def df2json(df: pd.DataFrame, head_tail_num=10):

    ret = None

    if len(df) > (head_tail_num * 2):
        ret = pd.concat([df.head(head_tail_num), df.tail(head_tail_num)])
    else:
        ret = df

    cmd = PandasCmd.from_df(df, ret)
    ret = flask.json.dumps(cmd, default=json_converter)
    return ret


def get_data_from_post():
    return json.loads(flask.request.get_data().decode())




import flask
from flask import request, Blueprint
import pandas as pd
from pdcopyist.src.core.Proxy import ProxyManager
from pdcopyist.src.controllers import common


app = Blueprint('table', __name__)


@app.route('/filters', methods=['post'])
def table_filters():
    param = common.get_data_from_post()

    px = ProxyManager.get()
    px.query(param['query'])
    return common.df2json(px.get_df_data())


@app.route('/handle', methods=['post'])
def table_handle():

    # todo:重构
    def _to_df_cols(df: pd.DataFrame, col: str):
        # 如果是多层索引，使用eval转成元组
        if isinstance(df.columns, pd.MultiIndex):
            return eval(col)

        return col

    param = common.get_data_from_post()

    filters = param['filters']
    sort = param['sort']

    filters = ((f['field'], f['values']) for f in filters)

    px = ProxyManager.get()
    px.filter(filters)

    df_ret = px.get_df_data()

    if sort:
        order = True if sort['order'] == 'asc' else False
        df_ret = df_ret.sort_values(_to_df_cols(
            df_ret, sort['field']), ascending=order)

    return common.df2json(df_ret)




import itertools
from typing import Generator, List, Iterable


class CodeFormatter(object):

    @staticmethod
    def type2str(v):
        if isinstance(v, str):
            return f'"{v}"' if "'" in v else f"'{v}'"

        return str(v)

    def to_code(self, method_name: str, args, kws) -> str:
        raise NotImplementedError


class StandardCodeFormatter(CodeFormatter):

    def convert_args_type(self, args: list) -> Generator[str, None, None]:
        return (
            CodeFormatter.type2str(a)
            for a in args
        )

    def convert_kws_type(self, kws: dict) -> Generator[str, None, None]:
        return (
            f'{k} = {CodeFormatter.type2str(v)}'
            for k, v in kws.items()
        )

    def do_to_code(self, method_name: str, args: Iterable[str], kws: Iterable[str]):
        raise NotImplementedError

    def to_code(self, method_name: str, args: list, kws: dict) -> str:
        args2type = self.convert_args_type(args)
        kws2type = self.convert_kws_type(kws)

        return self.do_to_code(method_name, args2type, kws2type)


class MethodCodeFormatter(StandardCodeFormatter):
    def do_to_code(self, method_name: str, args: Iterable[str], kws: Iterable[str]):
        smr_args = ','.join(itertools.chain(args, kws))
        return f'.{method_name}({smr_args})'


class GetItemCodeFormatter(StandardCodeFormatter):
    def do_to_code(self, method_name: str, args: Iterable[str], kws: Iterable[str]):
        smr_args = ','.join(itertools.chain(args, kws))
        return f'[{smr_args}]'


class OperatorCodeFormatter(StandardCodeFormatter):
    def do_to_code(self, operator: str, args: Iterable[str], kws: Iterable[str]):
        other = list(args)[0]
        return f'{operator}{other}'


class ImportCodeFormatter(CodeFormatter):

    pass


if __name__ == "__main__":
    pass



from pdcopyist.src.cusFuns.core.DecoratorFuns import dt_handle_func, dt_args, dt_source_code
from pdcopyist.src.cusFuns.core import Args as ty
import pandas as pd


def generate_code(*args, **kwargs):
    return f'''df.to_excel(r'{kwargs['path']}','{kwargs['sheet_name']}',index={kwargs['index']})'''


@dt_source_code(generate_code)
@dt_handle_func('导出文件-Excel')
@dt_args(path=ty.Input('文件全路径：'), sheet_name=ty.Input('工作表名字：'), index=ty.Switch('是否输出行索引：', default=False))
def remove_cols(df: pd.DataFrame, path, sheet_name, index):
    return df.to_excel(path, sheet_name)



from pdcopyist.src.cusFuns.core.DecoratorFuns import dt_handle_func, dt_args, dt_source_code
from pdcopyist.src.cusFuns.core import Args as ty
import pandas as pd


def generate_code(*args, **kwargs):
    return f'''df.to_feather(r'{kwargs['path']}')'''


@dt_source_code(generate_code)
@dt_handle_func('导出文件-feather')
@dt_args(path=ty.Input('文件全路径：'))
def remove_cols(df: pd.DataFrame, path):
    return df.to_feather(path)




from flask import Blueprint
from pdcopyist.src.core.Proxy import ProxyManager
from pdcopyist.src.controllers import common

app = Blueprint('groupby', __name__)


@app.route('/', methods=['post'])
def groupby():
    param = common.get_data_from_post()
    keys = param['_keys']
    aggs = param['_aggs']

    aggs = {d['filed']: d['methods'] for d in aggs}

    px = ProxyManager.get()
    px.groupby(keys, aggs)

    return common.df2json(px.get_df_data())



from typing import List

import uuid

from pdcopyist.src.core.Statement import AbcStatement, CodeStatement, Statement

from pdcopyist.src.helper.utils import CanJson
import itertools


class Commander(CanJson):

    def to_json_dict(self):
        ret = self.__dict__.copy()
        del ret['statements']
        ret.update({'hasError': self.has_exception()})
        ret.update({'code': self.to_code()})
        return ret

    def __init__(self, name: str) -> None:
        self._id = uuid.uuid1()
        self.name = name
        self.statements: List[AbcStatement] = []

    def __enter__(self) -> 'Commander':
        return self

    def __exit__(self, type, value, trace):
        pass

    @property
    def id(self):
        return self._id

    def create_statement(self, caller_name: str, ret_name: str = None) -> Statement:
        st = Statement(caller_name, ret_name)
        self.addStatement(st)
        return st

    def create_code_statement(self, code) -> CodeStatement:
        st = CodeStatement(code)
        self.addStatement(st)
        return st

    def get_last_statement(self) -> AbcStatement:
        return self.statements[-1]

    def has_exception(self):
        ret = any(s.has_exception() for s in self.statements)
        return ret

    def is_empty(self) -> bool:
        return len(self.statements) == 0

    def addStatement(self, statement: AbcStatement):
        self.statements.append(statement)

    def to_code(self) -> str:
        codes = (
            c.to_code()
            for c in self.statements
        )

        codes = itertools.chain([f'#{self.name}'], codes)

        code = '\n'.join(codes)

        return code


class CommanderManager():

    def __init__(self) -> None:
        self._cmds: List[Commander] = []
        self.reset_cmds()
        self.var_dict = {}

    def get_last_cmd(self) -> Commander:
        if len(self._cmds) == 0:
            return None
        return self._cmds[-1]

    def get_cmds(self):
        return (c for c in self._cmds if not c.is_empty())

    def addVar(self, var_name: str, ret):
        self.var_dict[var_name] = ret

    def getVar(self, var_name: str):
        return self.var_dict[var_name]

    def get_last_used_var_ret(self):
        key = list(self.var_dict.keys())[-1]
        return self.getVar(key)

    def clear_all_var(self):
        self.var_dict.clear()

    def clear_all(self):
        self._cmds.clear()
        self.clear_all_var()

    def append(self, cmd: Commander):
        self._cmds.append(cmd)

    def reset_cmds(self):
        self._cmds: List[Commander] = []

    def remove_cmd(self, id: uuid.UUID):
        idxs = [i for i, t in enumerate(self._cmds) if t.id == id]
        assert (len(idxs) in [0, 1]), f'多个 cmd[id={id}] 符合条件'
        if len(idxs) > 0:
            del self._cmds[idxs[0]]

    def to_code(self) -> str:
        codes = (
            f'{c.to_code()}'
            for c in self.get_cmds()
            if not c.is_empty()
        )

        return '\n\n'.join(codes)





from typing import Any, Callable, List

from pdcopyist.src.cusFuns.core.UIModl import FunModel

from pdcopyist.src.cusFuns.core.Args import AbcArgs


class CusFunsWrapper(object):
    def __init__(self, org_func) -> None:
        self.org_func = org_func


class HandleWrapper(CusFunsWrapper):

    @staticmethod
    def create_if_not(func) -> 'HandleWrapper':
        if not isinstance(func, HandleWrapper):
            func = HandleWrapper(func)
        return func

    def __init__(self, org_func) -> None:
        super().__init__(org_func)
        self.handle_func: Callable = None
        self.fun_name: str = None
        self.ui_json_func: Callable = None
        self.code_func: Callable = None
        self.args: List[AbcArgs] = []

    def auto_create_ui_func(self):
        if self.ui_json_func is not None:
            return

        def _ui_json_func() -> FunModel:
            fm = FunModel(self.fun_name)
            rules = {a.var_name: a.to_rule() for a in self.args}
            for a in self.args:
                fm.add_content(a.to_content())

            fm.rules = rules
            return fm

        self.ui_json_func = _ui_json_func


def dt_handle_func(fun_name=None, ui_json_func=None):

    def wrapper(func):
        func = HandleWrapper.create_if_not(func)

        func.handle_func = func.org_func
        func.fun_name = fun_name
        func.ui_json_func = ui_json_func
        return func

    return wrapper


def dt_args(**kwargs):

    def wrapper(func):
        func = HandleWrapper.create_if_not(func)

        func.args.extend((arg.set_var_name(var_name)
                          for var_name, arg in kwargs.items()))
        return func

    return wrapper


def dt_source_code(generate_func: Callable):
    def wrapper(func):
        func = HandleWrapper.create_if_not(func)
        func.code_func = generate_func

        return func

    return wrapper




import json
import flask
from flask import request, Blueprint
import pandas as pd
from pdcopyist.src.core.Proxy import ProxyManager
from pdcopyist.src.controllers import common

app = Blueprint('loaddata', __name__)


# @app.route('/get_file_args/ext=<ext>', methods=['get'])
# def get_file_args(ext: str):
#     print(ext)
#     print(request.values.get('ext'))
#     return 'done'


@app.route('/file', methods=['post'])
def upload_file():

    file = flask.request.files.get('file')

    ext = flask.request.form['ext']
    filename = flask.request.form['file_name']
    args = json.loads(flask.request.form['args'])

    # todo:根据不同情况做处理
    px = ProxyManager.get()
    px.read_data(file, filename, ext, args)
    return common.df2json(px.get_df_data())
